"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _JobStatus:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _JobStatusEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_JobStatus.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNKNOWN: _JobStatus.ValueType  # 0
    ERROR: _JobStatus.ValueType  # 1
    RECEIVED: _JobStatus.ValueType  # 2
    PRE_JOB: _JobStatus.ValueType  # 3
    READ_DB: _JobStatus.ValueType  # 4
    COMPUTE: _JobStatus.ValueType  # 5
    COMPLETED: _JobStatus.ValueType  # 6

class JobStatus(_JobStatus, metaclass=_JobStatusEnumTypeWrapper): ...

UNKNOWN: JobStatus.ValueType  # 0
ERROR: JobStatus.ValueType  # 1
RECEIVED: JobStatus.ValueType  # 2
PRE_JOB: JobStatus.ValueType  # 3
READ_DB: JobStatus.ValueType  # 4
COMPUTE: JobStatus.ValueType  # 5
COMPLETED: JobStatus.ValueType  # 6
global___JobStatus = JobStatus

class _ComputationMethod:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _ComputationMethodEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_ComputationMethod.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    COMPUTATION_METHOD_UNSPECIFIED: _ComputationMethod.ValueType  # 0
    COMPUTATION_METHOD_MEAN: _ComputationMethod.ValueType  # 1
    COMPUTATION_METHOD_VARIANCE: _ComputationMethod.ValueType  # 2
    COMPUTATION_METHOD_SUM: _ComputationMethod.ValueType  # 3
    COMPUTATION_METHOD_CORREL: _ComputationMethod.ValueType  # 4
    COMPUTATION_METHOD_MESH_CODE: _ComputationMethod.ValueType  # 5
    COMPUTATION_METHOD_JOIN_TABLE: _ComputationMethod.ValueType  # 6

class ComputationMethod(_ComputationMethod, metaclass=_ComputationMethodEnumTypeWrapper): ...

COMPUTATION_METHOD_UNSPECIFIED: ComputationMethod.ValueType  # 0
COMPUTATION_METHOD_MEAN: ComputationMethod.ValueType  # 1
COMPUTATION_METHOD_VARIANCE: ComputationMethod.ValueType  # 2
COMPUTATION_METHOD_SUM: ComputationMethod.ValueType  # 3
COMPUTATION_METHOD_CORREL: ComputationMethod.ValueType  # 4
COMPUTATION_METHOD_MESH_CODE: ComputationMethod.ValueType  # 5
COMPUTATION_METHOD_JOIN_TABLE: ComputationMethod.ValueType  # 6
global___ComputationMethod = ComputationMethod

class _ShareValueTypeEnum:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _ShareValueTypeEnumEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_ShareValueTypeEnum.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    SHARE_VALUE_TYPE_UNSPECIFIED: _ShareValueTypeEnum.ValueType  # 0
    SHARE_VALUE_TYPE_FIXED_POINT: _ShareValueTypeEnum.ValueType  # 1
    SHARE_VALUE_TYPE_UTF_8_INTEGER_REPRESENTATION: _ShareValueTypeEnum.ValueType  # 2

class ShareValueTypeEnum(_ShareValueTypeEnum, metaclass=_ShareValueTypeEnumEnumTypeWrapper): ...

SHARE_VALUE_TYPE_UNSPECIFIED: ShareValueTypeEnum.ValueType  # 0
SHARE_VALUE_TYPE_FIXED_POINT: ShareValueTypeEnum.ValueType  # 1
SHARE_VALUE_TYPE_UTF_8_INTEGER_REPRESENTATION: ShareValueTypeEnum.ValueType  # 2
global___ShareValueTypeEnum = ShareValueTypeEnum

class ProcedureProgress(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    PROGRESS_FIELD_NUMBER: builtins.int
    COMPLETED_FIELD_NUMBER: builtins.int
    DETAILS_FIELD_NUMBER: builtins.int
    id: builtins.int
    """unique id in ProgressManager context"""
    description: builtins.str
    """describe specific process in job"""
    progress: builtins.float
    """describe progress by [0.0, 100.0]"""
    completed: builtins.bool
    """describe that procedure is completed or not"""
    details: builtins.str
    """describe progress in details"""
    def __init__(
        self,
        *,
        id: builtins.int = ...,
        description: builtins.str = ...,
        progress: builtins.float = ...,
        completed: builtins.bool = ...,
        details: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_details", b"_details", "details", b"details"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_details", b"_details", "completed", b"completed", "description", b"description", "details", b"details", "id", b"id", "progress", b"progress"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_details", b"_details"]) -> typing_extensions.Literal["details"] | None: ...

global___ProcedureProgress = ProcedureProgress

class JobProgress(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOB_UUID_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    PROGRESSES_FIELD_NUMBER: builtins.int
    job_uuid: builtins.str
    status: global___JobStatus.ValueType
    @property
    def progresses(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ProcedureProgress]: ...
    def __init__(
        self,
        *,
        job_uuid: builtins.str = ...,
        status: global___JobStatus.ValueType = ...,
        progresses: collections.abc.Iterable[global___ProcedureProgress] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["job_uuid", b"job_uuid", "progresses", b"progresses", "status", b"status"]) -> None: ...

global___JobProgress = JobProgress

class Stacktrace(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class Frame(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        SOURCE_LOCATION_FIELD_NUMBER: builtins.int
        SOURCE_LINE_FIELD_NUMBER: builtins.int
        FUNCTION_NAME_FIELD_NUMBER: builtins.int
        source_location: builtins.str
        source_line: builtins.int
        function_name: builtins.str
        def __init__(
            self,
            *,
            source_location: builtins.str = ...,
            source_line: builtins.int = ...,
            function_name: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["function_name", b"function_name", "source_line", b"source_line", "source_location", b"source_location"]) -> None: ...

    FRAMES_FIELD_NUMBER: builtins.int
    @property
    def frames(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Stacktrace.Frame]: ...
    def __init__(
        self,
        *,
        frames: collections.abc.Iterable[global___Stacktrace.Frame] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["frames", b"frames"]) -> None: ...

global___Stacktrace = Stacktrace

class JobErrorInfo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    WHAT_FIELD_NUMBER: builtins.int
    ADDITIONAL_INFO_FIELD_NUMBER: builtins.int
    STACKTRACE_FIELD_NUMBER: builtins.int
    what: builtins.str
    additional_info: builtins.str
    @property
    def stacktrace(self) -> global___Stacktrace: ...
    def __init__(
        self,
        *,
        what: builtins.str = ...,
        additional_info: builtins.str | None = ...,
        stacktrace: global___Stacktrace | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_additional_info", b"_additional_info", "_stacktrace", b"_stacktrace", "additional_info", b"additional_info", "stacktrace", b"stacktrace"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_additional_info", b"_additional_info", "_stacktrace", b"_stacktrace", "additional_info", b"additional_info", "stacktrace", b"stacktrace", "what", b"what"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_additional_info", b"_additional_info"]) -> typing_extensions.Literal["additional_info"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_stacktrace", b"_stacktrace"]) -> typing_extensions.Literal["stacktrace"] | None: ...

global___JobErrorInfo = JobErrorInfo

class Schema(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    name: builtins.str
    type: global___ShareValueTypeEnum.ValueType
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        type: global___ShareValueTypeEnum.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name", "type", b"type"]) -> None: ...

global___Schema = Schema



from dataclasses import dataclass
from typing import Union

from .proto.common_types.common_types_pb2 import JobErrorInfo


@dataclass(frozen=True)
class QMPCJobError(Exception):
    err_info: JobErrorInfo


@dataclass(frozen=True)
class QMPCServerError(Exception):
    err_info: Union[str, JobErrorInfo]


class ArgumentError(Exception):
    pass



import math
from decimal import Decimal

import numpy as np
import pytest

from quickmpc.exception import ArgumentError
from quickmpc.share import Share


class TestQMPC:

    @staticmethod
    def isclose(val, true_val):
        ok: bool = True
        if isinstance(val, list):
            for v, t in zip(val, true_val):
                ok &= TestQMPC.isclose(v, t)
        elif isinstance(val, dict):
            for v, t in zip(val.values(), true_val.values()):
                ok &= TestQMPC.isclose(v, t)
        else:
            ok &= math.isclose(val, true_val)
        return ok

    @pytest.mark.parametrize(
        ("shares", "expected"),
        [
            # 1次元配列のシェア
            # int
            ([1, 2, 3], 6),
            # float
            ([2.4, 4.2, 4.1], 10.7),
            # Decimal
            ([Decimal(2.4), Decimal(4.2), Decimal(4.1)], Decimal(10.7)),

            # 2次元配列のシェア
            # int
            ([[1, 4], [2, 5], [3, 6]], [6, 15]),
            # float
            ([[1.1, 4.2], [2.3, 5.4], [3.5, 6.6]], [6.9, 16.2]),
            # Decimal
            ([[Decimal(1.1), Decimal(4.2)], [Decimal(2.3), Decimal(5.4)],
              [Decimal(3.5), Decimal(6.6)]], [Decimal(6.9), Decimal(16.2)]),

            # 3次元配列のシェア
            # int
            ([[[1, 4], [7, 8]],
              [[2, 5], [9, 10]],
              [[3, 6], [11, 12]]],
                [[6, 15], [27, 30]]),
            # float
            ([[[1.1, 4.2], [7.1, 8.2]],
              [[2.3, 5.4], [9.1, 10.2]],
              [[3.5, 6.6], [11.1, 12.2]]],
             [[6.9, 16.2], [27.3, 30.6]]),
            # Decimal
            ([[[Decimal(1.1), Decimal(4.2)], [Decimal(7.1), (8.2)]],
              [[Decimal(2.3), Decimal(5.4)], [Decimal(9.1), (10.2)]],
              [[Decimal(3.5), Decimal(6.6)], [Decimal(11.1), (12.2)]]],
             [[Decimal(6.9), Decimal(16.2)], [Decimal(27.3), (30.6)]]),

        ]
    )
    def test_recons_list(self, shares: list, expected):
        """ 3パーティの復元が正しくできるかTest """
        secrets: list = Share.recons(shares)
        assert TestQMPC.isclose(secrets, expected)

    @pytest.mark.parametrize(
        ("shares", "expected"),
        [
            # 辞書型配列のシェア
            ([{"a": 1}, {"a": 2}, {"a": 3}],
             {"a": 6}),
            ([{"a": 1.1, "b": 2}, {"a": 4.1, "b": 9.1}, {"a": 3.1, "b": -2.1}],
             {"a": 8.3, "b": 9}),
            ([{"a": 1.1, "child": {"b": 2, "c": [1, 3]}},
              {"a": 4.1, "child": {"b": 9.1, "c": [2, 1]}},
              {"a": 3.1, "child": {"b": -2.1, "c": [1, 8]}}],
             {"a": 8.3, "child": {"b": 9, "c": [4, 12]}}),
            # 辞書型配列の配列のシェア
            ([[{"a": 1}], [{"a": 2}], [{"a": 3}]],
             [{"a": 6}]),
            ([[{"a": 1.1}, {"b": 2}],
                [{"a": 3.1}, {"b": 9.1}],
                [{"a": 4.1}, {"b": -2.1}]],
                [{"a": 8.3}, {"b": 9}]),
            ([[{"a": 1.1}, {"child": {"b": 2, "c": [1, 3]}}],
                [{"a": 4.1}, {"child": {"b": 9.1, "c": [2, 1]}}],
                [{"a": 3.1}, {"child": {"b": -2.1, "c": [1, 8]}}]],
                [{"a": 8.3}, {"child": {"b": 9, "c": [4, 12]}}]),
        ]
    )
    def test_recons_dict(self, shares: dict, expected: dict):
        """ 3パーティの辞書型の復元が正しくできるかTest """
        secrets: list = Share.recons(shares)
        assert TestQMPC.isclose(secrets, expected)

    @pytest.mark.parametrize(
        ("shares", "expected"),
        [
            # 2パーティのシェア
            ([[1, 5], [2, 9]], [3, 14]),
            ([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], [[6, 8], [10, 12]]),
            ([[2.4, 9.1], [4.2, 6.7], [4.1, 1.5]], [10.7, 17.3]),
            # 4パーティのシェア
            ([[3000], [2500], [200], [400]], [6100]),
            ([[[3000]], [[2500]], [[200]], [[400]]], [[6100]]),
            # 5パーティのシェア
            ([[3000], [2500], [200], [400], [11235]], [17335]),
            ([[[3000]], [[2500]], [[200]], [[400]], [[11235]]], [[17335]]),
        ]
    )
    def test_recons_multi(self, shares: list, expected: list):
        """ nパーティのシェアの復元が正しくできるかTest """
        secrets: list = Share.recons(shares)
        assert TestQMPC.isclose(secrets, expected)

    @pytest.mark.parametrize(
        ("shares", "expected"),
        [
            ([1e1200 for _ in range(3)], 3e1200),
            ([Decimal(1e1200) for _ in range(3)], Decimal(3e1200)),
            ([-1e1200 for _ in range(3)], -3e1200),
            ([Decimal(-1e1200) for _ in range(3)], Decimal(-3e1200)),
            ([1e-50 for _ in range(3)], 3e-50),
            ([Decimal(1e-50) for _ in range(3)], Decimal(3e-50)),
            ([-1e-50 for _ in range(3)], -3e-50),
            ([Decimal(-1e-50) for _ in range(3)], Decimal(-3e-50)),
        ]
    )
    def test_recons_edge(self, shares: list, expected):
        """ 3パーティの復元が正しくできるかTest """
        secrets = Share.recons(shares)
        assert TestQMPC.isclose(secrets, expected)

    @pytest.mark.parametrize(
        ("shares", "expected"),
        [
            # scalar value is not allowed
            (1, ArgumentError),

            # string value is not allowed
            ("hey", ArgumentError),
            (["hey"], TypeError),  # TODO: quickmpcのArgumentErrorが出るようにしたい
            ([["hey"]], TypeError),
        ]
    )
    def test_recons_errorhandring(self, shares, expected):
        """ 異常値を与えてエラーが出るかTest """
        with pytest.raises(expected):
            Share.recons(shares)

    @pytest.mark.parametrize(
        ("secrets"),
        [
            # 1次元配列の秘密情報
            ([1]), ([3.0, 5.12, 21.05]), ([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
            # 2次元配列の秘密情報
            ([[1]]), ([[3000, 500, 2100], [100, 2500, 1100]]), ([[2.1, 3.2]]),
            ([[1, 2, 3, 4, 5],
              [6, 7, 8, 9, 10],
              [11, 12, 13, 14, 15],
              [16, 17, 18, 19, 20],
              [21, 22, 23, 24, 25]])
        ]
    )
    def test_sharize_recons(self, secrets: list):
        """ nパーティのシェア化，復元が正しくできてるかTest"""
        for party_size in range(2, 10):
            shares: list = Share.sharize(secrets, party_size=party_size)
            conved: list = [Share.convert_type(s) for s in shares]
            secrets_2: list = Share.recons(conved)
            secrets_2 = Share.convert_type(secrets_2)
            assert (np.allclose(secrets, secrets_2))



import logging
from dataclasses import dataclass
from decimal import Decimal
from typing import (Any, Callable, ClassVar, List,
                    Optional, Sequence, Tuple, Union)

import numpy as np

from .exception import ArgumentError
from .proto.common_types.common_types_pb2 import Schema, ShareValueTypeEnum
from .utils.overload_tools import (DictList, DictList2, Dim1,
                                   Dim2, Dim3, methoddispatch)
from .utils.random import ChaCha20, RandomInterface

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Share:
    __share_random_range: ClassVar[Tuple[Decimal, Decimal]] =\
        (Decimal(-(1 << 64)), Decimal(1 << 64))

    @methoddispatch(is_static_method=True)
    @staticmethod
    def __to_str(_):
        logger.error("Invalid argument on stringfy.")
        raise ArgumentError("不正な引数が与えられています．")

    @__to_str.register(Decimal)
    @staticmethod
    def __decimal_to_str(val: Decimal) -> str:
        # InfinityをCCで読み込めるinfに変換
        return 'inf' if Decimal.is_infinite(val) else str(val)

    @__to_str.register(int)
    @staticmethod
    def __int_to_str(val: int) -> str:
        return str(val)

    @methoddispatch(is_static_method=True)
    @staticmethod
    def sharize(_, __):
        logger.error("Invalid argument on sharize.")
        raise ArgumentError("不正な引数が与えられています．")

    @methoddispatch(is_static_method=True)
    @staticmethod
    def recons(_):
        logger.error("Invalid argument on recons.")
        raise ArgumentError("不正な引数が与えられています．")

    @methoddispatch(is_static_method=True)
    @staticmethod
    def convert_type(_, __):
        logger.error("Invalid argument on convert_type.")
        raise ArgumentError("不正な引数が与えられています．")

    @sharize.register(int)
    @sharize.register(float)
    @staticmethod
    def __sharize_scalar(secrets: float, party_size: int = 3) -> List[str]:
        """ スカラ値のシェア化 """
        rnd: RandomInterface = ChaCha20()
        shares: List[int] = rnd.get_list(
            *Share.__share_random_range, party_size)
        shares[0] += Decimal(secrets) - np.sum(shares)
        shares_str: List[str] = [str(n) for n in shares]
        return shares_str

    @sharize.register((Dim1, float))
    @staticmethod
    def __sharize_1dimension_float(secrets: List[Union[float, Decimal]],
                                   party_size: int = 3) \
            -> List[List[str]]:
        """ 1次元リストのシェア化 """
        rnd: RandomInterface = ChaCha20()
        secrets_size: int = len(secrets)
        shares: np.ndarray = np.array([
            rnd.get_list(*Share.__share_random_range, secrets_size)
            for __ in range(party_size - 1)])
        s1: np.ndarray = np.subtract(np.frompyfunc(Decimal, 1, 1)(secrets),
                                     np.sum(shares, axis=0))
        shares_str: List[List[str]] = \
            np.vectorize(Share.__to_str)([s1, *shares]).tolist()
        return shares_str

    @sharize.register((Dim1, Decimal))
    @staticmethod
    def __sharize_1dimension_decimal(secrets: List[Decimal],
                                     party_size: int = 3) \
            -> List[List[str]]:
        return Share.__sharize_1dimension_float(secrets, party_size)

    @sharize.register((Dim1, int))
    @staticmethod
    def __sharize_1dimension_int(secrets: List[int], party_size: int = 3) \
            -> List[List[str]]:
        """ 1次元リストのシェア化 """
        rnd: RandomInterface = ChaCha20()
        secrets_size: int = len(secrets)
        max_val = (max(secrets)+1) * 2
        shares: np.ndarray = np.array([
            rnd.get_list(-max_val, max_val, secrets_size)
            for __ in range(party_size - 1)])
        s1: np.ndarray = np.subtract(np.frompyfunc(int, 1, 1)(secrets),
                                     np.sum(shares, axis=0))
        shares_str: List[List[str]] = np.vectorize(
            Share.__to_str)([s1, *shares]).tolist()
        return shares_str

    @sharize.register(Dim2)
    @staticmethod
    def __sharize_2dimension(secrets: List[List[Union[float, int]]],
                             party_size: int = 3) -> List[List[List[str]]]:
        """ 2次元リストのシェア化 """
        transposed: List[Union[List[int], List[float]]] \
            = np.array(secrets, dtype=object).transpose().tolist()
        dst: List[List[List[str]]] = [
            Share.sharize(col, party_size) for col in transposed
        ]
        dst = np.array(dst, dtype=object).transpose(1, 2, 0).tolist()

        return dst

    @sharize.register(dict)
    @staticmethod
    def __sharize_dict(secrets: dict, party_size: int = 3) -> List[dict]:
        """ 辞書型のシェア化 """
        shares_str: List[dict] = [dict() for _ in range(party_size)]
        for key, val in secrets.items():
            for i, share_val in enumerate(Share.sharize(val, party_size)):
                shares_str[i][key] = share_val
        return shares_str

    @sharize.register(DictList)
    @staticmethod
    def __sharize_dictlist(secrets: dict, party_size: int = 3) \
            -> List[List[dict]]:
        """ 辞書型配列のシェア化 """
        shares_str: List[List[dict]] = [[] for _ in range(party_size)]
        for secret_dict in secrets:
            share_dict: List[dict] = Share.sharize(secret_dict, party_size)
            for ss, sd in zip(shares_str, share_dict):
                ss.append(sd)
        return shares_str

    @recons.register(Dim1)
    @staticmethod
    def __recons_list1(shares: List[Union[int, Decimal]]):
        """ 1次元リストのシェアを復元 """
        return sum(shares)

    @recons.register(Dim2)
    @recons.register(Dim3)
    @staticmethod
    def __recons_list(shares: List[List[Union[int, Decimal]]]) -> List:
        """ リストのシェアを復元 """
        secrets: List = [
            Share.recons([shares_pi[i] for shares_pi in shares])
            for i in range(len(shares[0]))
        ]
        return secrets

    @recons.register(DictList)
    @staticmethod
    def __recons_dictlist(shares: List[dict]) -> dict:
        """ 辞書型を復元 """
        secrets: dict = dict()
        for key in shares[0].keys():
            val = []
            for s in shares:
                val.append(s[key])
            secrets[key] = Share.recons(val)
        return secrets

    @recons.register(DictList2)
    @staticmethod
    def __recons_dictlist2(shares: List[List[dict]]) -> list:
        """ 辞書型配列を復元 """
        secrets: list = list()
        for i in range(len(shares[0])):
            val = []
            for s in shares:
                val.append(s[i])
            secrets.append(Share.recons(val))
        return secrets

    @staticmethod
    def get_pre_convert_func(
            schema: Optional[Schema]) -> Callable[[str], Any]:
        """ スキーマに合った変換関数を返す  """
        if schema is None:
            return Decimal
        type = schema.type
        if type == ShareValueTypeEnum.Value('SHARE_VALUE_TYPE_UNSPECIFIED'):
            return Decimal
        if type == ShareValueTypeEnum.Value('SHARE_VALUE_TYPE_FIXED_POINT'):
            return Decimal
        if type == ShareValueTypeEnum.Value(
                'SHARE_VALUE_TYPE_UTF_8_INTEGER_REPRESENTATION'):
            # TODO: Decimal の経由を不要にする
            return lambda x: int(Decimal(x))
        return float

    @staticmethod
    def convert_int_to_str(x: int):
        bytes_repr: bytes = x.to_bytes((x.bit_length() + 7) //
                                       8, byteorder='big')
        str_repr: str = bytes_repr.decode('utf-8')
        return str_repr

    @staticmethod
    def get_convert_func(
            schema: Optional[Schema]) -> Callable[[Any], Any]:
        """ スキーマに合った変換関数を返す  """
        if schema is None:
            return float
        type = schema.type
        if type == ShareValueTypeEnum.Value('SHARE_VALUE_TYPE_UNSPECIFIED'):
            return float
        if type == ShareValueTypeEnum.Value('SHARE_VALUE_TYPE_FIXED_POINT'):
            return float
        if type == ShareValueTypeEnum.Value(
                'SHARE_VALUE_TYPE_UTF_8_INTEGER_REPRESENTATION'):
            return Share.convert_int_to_str
        return float

    @convert_type.register(str)
    @staticmethod
    def __pre_convert_type_str(
            value: str, schema: Optional[Schema] = None) -> list:
        func = Share.get_pre_convert_func(schema)
        return func(value)

    @convert_type.register(Dim1)
    @staticmethod
    def __convert_type_list(
            values: List[Any],
            schema: Optional[Sequence[Optional[Schema]]] = None) -> list:
        if schema is None:
            schema = [None] * len(values)
        return [Share.convert_type(x, sch)
                for x, sch in zip(values, schema)]

    @convert_type.register(Decimal)
    @convert_type.register(int)
    @staticmethod
    def __convert_type_elem(
            value: Union[Decimal, int],
            schema: Optional[Schema] = None) -> Union[float, str]:
        func = Share.get_convert_func(schema)
        return func(value)

    @convert_type.register(Dim2)
    @staticmethod
    def __convert_type_table(
            table: List[List],
            schema: Optional[List[Schema]] = None) -> list:
        return [Share.convert_type(row, schema) for row in table]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import collections.abc
import grpc
import libc_to_manage_pb2

class LibcToManageStub:
    """*
    LibcToManage service
    """

    def __init__(self, channel: grpc.Channel) -> None: ...
    SendShares: grpc.UnaryUnaryMultiCallable[
        libc_to_manage_pb2.SendSharesRequest,
        libc_to_manage_pb2.SendSharesResponse,
    ]
    DeleteShares: grpc.UnaryUnaryMultiCallable[
        libc_to_manage_pb2.DeleteSharesRequest,
        libc_to_manage_pb2.DeleteSharesResponse,
    ]
    GetSchema: grpc.UnaryUnaryMultiCallable[
        libc_to_manage_pb2.GetSchemaRequest,
        libc_to_manage_pb2.GetSchemaResponse,
    ]
    ExecuteComputation: grpc.UnaryUnaryMultiCallable[
        libc_to_manage_pb2.ExecuteComputationRequest,
        libc_to_manage_pb2.ExecuteComputationResponse,
    ]
    GetComputationResult: grpc.UnaryStreamMultiCallable[
        libc_to_manage_pb2.GetComputationResultRequest,
        libc_to_manage_pb2.GetComputationResultResponse,
    ]
    GetDataList: grpc.UnaryUnaryMultiCallable[
        libc_to_manage_pb2.GetDataListRequest,
        libc_to_manage_pb2.GetDataListResponse,
    ]
    GetElapsedTime: grpc.UnaryUnaryMultiCallable[
        libc_to_manage_pb2.GetElapsedTimeRequest,
        libc_to_manage_pb2.GetElapsedTimeResponse,
    ]
    GetJobErrorInfo: grpc.UnaryUnaryMultiCallable[
        libc_to_manage_pb2.GetJobErrorInfoRequest,
        libc_to_manage_pb2.GetJobErrorInfoResponse,
    ]

class LibcToManageServicer(metaclass=abc.ABCMeta):
    """*
    LibcToManage service
    """

    @abc.abstractmethod
    def SendShares(
        self,
        request: libc_to_manage_pb2.SendSharesRequest,
        context: grpc.ServicerContext,
    ) -> libc_to_manage_pb2.SendSharesResponse: ...
    @abc.abstractmethod
    def DeleteShares(
        self,
        request: libc_to_manage_pb2.DeleteSharesRequest,
        context: grpc.ServicerContext,
    ) -> libc_to_manage_pb2.DeleteSharesResponse: ...
    @abc.abstractmethod
    def GetSchema(
        self,
        request: libc_to_manage_pb2.GetSchemaRequest,
        context: grpc.ServicerContext,
    ) -> libc_to_manage_pb2.GetSchemaResponse: ...
    @abc.abstractmethod
    def ExecuteComputation(
        self,
        request: libc_to_manage_pb2.ExecuteComputationRequest,
        context: grpc.ServicerContext,
    ) -> libc_to_manage_pb2.ExecuteComputationResponse: ...
    @abc.abstractmethod
    def GetComputationResult(
        self,
        request: libc_to_manage_pb2.GetComputationResultRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[libc_to_manage_pb2.GetComputationResultResponse]: ...
    @abc.abstractmethod
    def GetDataList(
        self,
        request: libc_to_manage_pb2.GetDataListRequest,
        context: grpc.ServicerContext,
    ) -> libc_to_manage_pb2.GetDataListResponse: ...
    @abc.abstractmethod
    def GetElapsedTime(
        self,
        request: libc_to_manage_pb2.GetElapsedTimeRequest,
        context: grpc.ServicerContext,
    ) -> libc_to_manage_pb2.GetElapsedTimeResponse: ...
    @abc.abstractmethod
    def GetJobErrorInfo(
        self,
        request: libc_to_manage_pb2.GetJobErrorInfoRequest,
        context: grpc.ServicerContext,
    ) -> libc_to_manage_pb2.GetJobErrorInfoResponse: ...

def add_LibcToManageServicer_to_server(servicer: LibcToManageServicer, server: grpc.Server) -> None: ...



from quickmpc.qmpc import (
    QMPC, JobStatus, JobErrorInfo,
    Schema, ShareValueTypeEnum
)
from .version import __version__

__all__ = [
    "QMPC",
    "JobStatus",
    "JobErrorInfo",
    "Schema",
    "ShareValueTypeEnum",
    "__version__",
]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import common_types.common_types_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class SendSharesRequest(google.protobuf.message.Message):
    """*
    the message of SendSharesRequest
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATA_ID_FIELD_NUMBER: builtins.int
    SHARES_FIELD_NUMBER: builtins.int
    SCHEMA_FIELD_NUMBER: builtins.int
    PIECE_ID_FIELD_NUMBER: builtins.int
    SENT_AT_FIELD_NUMBER: builtins.int
    MATCHING_COLUMN_FIELD_NUMBER: builtins.int
    TOKEN_FIELD_NUMBER: builtins.int
    data_id: builtins.str
    shares: builtins.str
    @property
    def schema(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[common_types.common_types_pb2.Schema]: ...
    piece_id: builtins.int
    sent_at: builtins.str
    matching_column: builtins.int
    token: builtins.str
    def __init__(
        self,
        *,
        data_id: builtins.str = ...,
        shares: builtins.str = ...,
        schema: collections.abc.Iterable[common_types.common_types_pb2.Schema] | None = ...,
        piece_id: builtins.int = ...,
        sent_at: builtins.str = ...,
        matching_column: builtins.int = ...,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["data_id", b"data_id", "matching_column", b"matching_column", "piece_id", b"piece_id", "schema", b"schema", "sent_at", b"sent_at", "shares", b"shares", "token", b"token"]) -> None: ...

global___SendSharesRequest = SendSharesRequest

class SendSharesResponse(google.protobuf.message.Message):
    """*
    the message of SendSharesResponse
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGE_FIELD_NUMBER: builtins.int
    IS_OK_FIELD_NUMBER: builtins.int
    message: builtins.str
    is_ok: builtins.bool
    def __init__(
        self,
        *,
        message: builtins.str = ...,
        is_ok: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["is_ok", b"is_ok", "message", b"message"]) -> None: ...

global___SendSharesResponse = SendSharesResponse

class DeleteSharesRequest(google.protobuf.message.Message):
    """*
    the message of DeleteSharesRequest
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATAIDS_FIELD_NUMBER: builtins.int
    TOKEN_FIELD_NUMBER: builtins.int
    @property
    def dataIds(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    token: builtins.str
    def __init__(
        self,
        *,
        dataIds: collections.abc.Iterable[builtins.str] | None = ...,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["dataIds", b"dataIds", "token", b"token"]) -> None: ...

global___DeleteSharesRequest = DeleteSharesRequest

class DeleteSharesResponse(google.protobuf.message.Message):
    """*
    the message of DeleteSharesResponse
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGE_FIELD_NUMBER: builtins.int
    IS_OK_FIELD_NUMBER: builtins.int
    message: builtins.str
    is_ok: builtins.bool
    def __init__(
        self,
        *,
        message: builtins.str = ...,
        is_ok: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["is_ok", b"is_ok", "message", b"message"]) -> None: ...

global___DeleteSharesResponse = DeleteSharesResponse

class GetSchemaRequest(google.protobuf.message.Message):
    """*
    the message of GetSchemaRequest
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATA_ID_FIELD_NUMBER: builtins.int
    TOKEN_FIELD_NUMBER: builtins.int
    data_id: builtins.str
    token: builtins.str
    def __init__(
        self,
        *,
        data_id: builtins.str = ...,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["data_id", b"data_id", "token", b"token"]) -> None: ...

global___GetSchemaRequest = GetSchemaRequest

class GetSchemaResponse(google.protobuf.message.Message):
    """*
    the message of GetSchemaResponse
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGE_FIELD_NUMBER: builtins.int
    IS_OK_FIELD_NUMBER: builtins.int
    SCHEMA_FIELD_NUMBER: builtins.int
    message: builtins.str
    is_ok: builtins.bool
    @property
    def schema(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[common_types.common_types_pb2.Schema]: ...
    def __init__(
        self,
        *,
        message: builtins.str = ...,
        is_ok: builtins.bool = ...,
        schema: collections.abc.Iterable[common_types.common_types_pb2.Schema] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["is_ok", b"is_ok", "message", b"message", "schema", b"schema"]) -> None: ...

global___GetSchemaResponse = GetSchemaResponse

class JoinOrder(google.protobuf.message.Message):
    """*
    the message of ExecuteComputationRequest
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATAIDS_FIELD_NUMBER: builtins.int
    JOIN_FIELD_NUMBER: builtins.int
    INDEX_FIELD_NUMBER: builtins.int
    @property
    def dataIds(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def join(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def index(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        dataIds: collections.abc.Iterable[builtins.str] | None = ...,
        join: collections.abc.Iterable[builtins.int] | None = ...,
        index: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["dataIds", b"dataIds", "index", b"index", "join", b"join"]) -> None: ...

global___JoinOrder = JoinOrder

class Input(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SRC_FIELD_NUMBER: builtins.int
    TARGET_FIELD_NUMBER: builtins.int
    @property
    def src(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def target(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        src: collections.abc.Iterable[builtins.int] | None = ...,
        target: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["src", b"src", "target", b"target"]) -> None: ...

global___Input = Input

class ExecuteComputationRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    METHOD_ID_FIELD_NUMBER: builtins.int
    TOKEN_FIELD_NUMBER: builtins.int
    TABLE_FIELD_NUMBER: builtins.int
    ARG_FIELD_NUMBER: builtins.int
    method_id: common_types.common_types_pb2.ComputationMethod.ValueType
    token: builtins.str
    @property
    def table(self) -> global___JoinOrder: ...
    @property
    def arg(self) -> global___Input: ...
    def __init__(
        self,
        *,
        method_id: common_types.common_types_pb2.ComputationMethod.ValueType = ...,
        token: builtins.str = ...,
        table: global___JoinOrder | None = ...,
        arg: global___Input | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["arg", b"arg", "table", b"table"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["arg", b"arg", "method_id", b"method_id", "table", b"table", "token", b"token"]) -> None: ...

global___ExecuteComputationRequest = ExecuteComputationRequest

class ExecuteComputationResponse(google.protobuf.message.Message):
    """
    the message of ExecuteComputationResponse
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGE_FIELD_NUMBER: builtins.int
    IS_OK_FIELD_NUMBER: builtins.int
    JOB_UUID_FIELD_NUMBER: builtins.int
    message: builtins.str
    is_ok: builtins.bool
    job_uuid: builtins.str
    def __init__(
        self,
        *,
        message: builtins.str = ...,
        is_ok: builtins.bool = ...,
        job_uuid: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["is_ok", b"is_ok", "job_uuid", b"job_uuid", "message", b"message"]) -> None: ...

global___ExecuteComputationResponse = ExecuteComputationResponse

class GetComputationResultRequest(google.protobuf.message.Message):
    """*
    the message of GetComputationResultRequest
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOB_UUID_FIELD_NUMBER: builtins.int
    TOKEN_FIELD_NUMBER: builtins.int
    job_uuid: builtins.str
    token: builtins.str
    def __init__(
        self,
        *,
        job_uuid: builtins.str = ...,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["job_uuid", b"job_uuid", "token", b"token"]) -> None: ...

global___GetComputationResultRequest = GetComputationResultRequest

class GetComputationResultResponse(google.protobuf.message.Message):
    """*
    the message of GetComputationResultResponse
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGE_FIELD_NUMBER: builtins.int
    IS_OK_FIELD_NUMBER: builtins.int
    RESULT_FIELD_NUMBER: builtins.int
    COLUMN_NUMBER_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    PIECE_ID_FIELD_NUMBER: builtins.int
    PROGRESS_FIELD_NUMBER: builtins.int
    IS_DIM1_FIELD_NUMBER: builtins.int
    IS_DIM2_FIELD_NUMBER: builtins.int
    IS_SCHEMA_FIELD_NUMBER: builtins.int
    message: builtins.str
    is_ok: builtins.bool
    @property
    def result(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    column_number: builtins.int
    status: common_types.common_types_pb2.JobStatus.ValueType
    piece_id: builtins.int
    @property
    def progress(self) -> common_types.common_types_pb2.JobProgress: ...
    is_dim1: builtins.bool
    is_dim2: builtins.bool
    is_schema: builtins.bool
    def __init__(
        self,
        *,
        message: builtins.str = ...,
        is_ok: builtins.bool = ...,
        result: collections.abc.Iterable[builtins.str] | None = ...,
        column_number: builtins.int = ...,
        status: common_types.common_types_pb2.JobStatus.ValueType = ...,
        piece_id: builtins.int = ...,
        progress: common_types.common_types_pb2.JobProgress | None = ...,
        is_dim1: builtins.bool = ...,
        is_dim2: builtins.bool = ...,
        is_schema: builtins.bool = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_progress", b"_progress", "is_dim1", b"is_dim1", "is_dim2", b"is_dim2", "is_schema", b"is_schema", "progress", b"progress", "result_type", b"result_type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_progress", b"_progress", "column_number", b"column_number", "is_dim1", b"is_dim1", "is_dim2", b"is_dim2", "is_ok", b"is_ok", "is_schema", b"is_schema", "message", b"message", "piece_id", b"piece_id", "progress", b"progress", "result", b"result", "result_type", b"result_type", "status", b"status"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_progress", b"_progress"]) -> typing_extensions.Literal["progress"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["result_type", b"result_type"]) -> typing_extensions.Literal["is_dim1", "is_dim2", "is_schema"] | None: ...

global___GetComputationResultResponse = GetComputationResultResponse

class GetDataListRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKEN_FIELD_NUMBER: builtins.int
    token: builtins.str
    def __init__(
        self,
        *,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["token", b"token"]) -> None: ...

global___GetDataListRequest = GetDataListRequest

class GetDataListResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RESULT_FIELD_NUMBER: builtins.int
    IS_OK_FIELD_NUMBER: builtins.int
    result: builtins.str
    is_ok: builtins.bool
    def __init__(
        self,
        *,
        result: builtins.str = ...,
        is_ok: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["is_ok", b"is_ok", "result", b"result"]) -> None: ...

global___GetDataListResponse = GetDataListResponse

class GetElapsedTimeRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOB_UUID_FIELD_NUMBER: builtins.int
    TOKEN_FIELD_NUMBER: builtins.int
    job_uuid: builtins.str
    token: builtins.str
    def __init__(
        self,
        *,
        job_uuid: builtins.str = ...,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["job_uuid", b"job_uuid", "token", b"token"]) -> None: ...

global___GetElapsedTimeRequest = GetElapsedTimeRequest

class GetElapsedTimeResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IS_OK_FIELD_NUMBER: builtins.int
    ELAPSED_TIME_FIELD_NUMBER: builtins.int
    is_ok: builtins.bool
    elapsed_time: builtins.float
    def __init__(
        self,
        *,
        is_ok: builtins.bool = ...,
        elapsed_time: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["elapsed_time", b"elapsed_time", "is_ok", b"is_ok"]) -> None: ...

global___GetElapsedTimeResponse = GetElapsedTimeResponse

class GetJobErrorInfoRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOB_UUID_FIELD_NUMBER: builtins.int
    TOKEN_FIELD_NUMBER: builtins.int
    job_uuid: builtins.str
    token: builtins.str
    def __init__(
        self,
        *,
        job_uuid: builtins.str = ...,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["job_uuid", b"job_uuid", "token", b"token"]) -> None: ...

global___GetJobErrorInfoRequest = GetJobErrorInfoRequest

class GetJobErrorInfoResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOB_ERROR_INFO_FIELD_NUMBER: builtins.int
    IS_OK_FIELD_NUMBER: builtins.int
    @property
    def job_error_info(self) -> common_types.common_types_pb2.JobErrorInfo: ...
    is_ok: builtins.bool
    def __init__(
        self,
        *,
        job_error_info: common_types.common_types_pb2.JobErrorInfo | None = ...,
        is_ok: builtins.bool = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_job_error_info", b"_job_error_info", "job_error_info", b"job_error_info"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_job_error_info", b"_job_error_info", "is_ok", b"is_ok", "job_error_info", b"job_error_info"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_job_error_info", b"_job_error_info"]) -> typing_extensions.Literal["job_error_info"] | None: ...

global___GetJobErrorInfoResponse = GetJobErrorInfoResponse



"""
listの次元によるオーバロードが可能となるような自作dispatch
"""

from functools import update_wrapper


class Dim1:
    ...


class Dim2:
    ...


class Dim3:
    ...


class DictList:
    ...


class DictList2:
    ...


def d(lst):
    if not isinstance(lst, list):
        return 0
    if not lst:
        return 1
    return 1 + d(lst[0])


def find_element_type(lst) -> type:
    if not isinstance(lst, list):
        return lst.__class__
    if not lst:
        return None.__class__
    return find_element_type(lst[0])


def _get_dim_class(lst: list):
    dim: int = d(lst)
    if dim == 1:
        if len(lst) and isinstance(lst[0], dict):
            return DictList()
        return Dim1()
    elif dim == 2:
        if len(lst) and len(lst[0]) and isinstance(lst[0][0], dict):
            return DictList2()
        return Dim2()
    elif dim == 3:
        return Dim3()
    return None


def _convert_list_type(registry, type):
    if type.__class__ is not list:
        return type
    for Dim in (Dim1, Dim2, Dim3, DictList):
        if Dim in registry:
            return _get_dim_class(type)
    return type


def methoddispatch(is_static_method: bool = False):
    def _dimdispatch(func):
        registry: dict = {}
        default_function = func

        def wrapper(*args, **kw):
            arg = args[0] if is_static_method else args[1]
            type = _convert_list_type(registry, arg)
            elem_type = find_element_type(arg)
            key = (type.__class__, elem_type)
            func = default_function
            if type.__class__ in registry:
                func = registry[type.__class__]
            if key in registry:
                func = registry[key]
            if is_static_method:
                return func.__func__(*args, **kw)
            else:
                return func(*args, **kw)

        def register(types, func=None):
            if func is None:
                return lambda f: register(types, f)
            registry[types] = func
            return func

        wrapper.register = register
        update_wrapper(wrapper, func)
        return wrapper
    return _dimdispatch



import csv
import logging
from dataclasses import dataclass
from hashlib import sha512
from typing import Dict, List, Optional, Sequence, Tuple, Union

import numpy as np

from ..exception import ArgumentError
from ..proto.common_types.common_types_pb2 import Schema, ShareValueTypeEnum
from .overload_tools import Dim1, methoddispatch

logger = logging.getLogger(__name__)

SUPPORT_TAGS: Dict[str, ShareValueTypeEnum.ValueType] = {
    'id': ShareValueTypeEnum.Value('SHARE_VALUE_TYPE_FIXED_POINT'),
}

ShareValueType = Union[float, int]


@dataclass(frozen=True)
class FormatChecker:
    @methoddispatch(is_static_method=True)
    @staticmethod
    def check_duplicate(_):
        raise ArgumentError("不正な引数が与えられています．")

    @check_duplicate.register(Dim1)
    @staticmethod
    def check_duplicate_dummy(schema: List[str]):
        raise ArgumentError("不正な引数が与えられています．")

    @check_duplicate.register((Dim1, str))
    @staticmethod
    def check_duplicate_strs(schema: List[str]) -> bool:
        return len(schema) == len(set(schema))

    @check_duplicate.register((Dim1, Schema))
    @staticmethod
    def check_duplicate_typed(schema: List[Schema]) -> bool:
        return len(schema) == len(set([sch.name for sch in schema]))

    @staticmethod
    def check_size(secrets: Sequence[Sequence[Union[str, ShareValueType]]],
                   schema: Sequence[Union[str, Schema]]) -> np.bool_:
        return np.all([len(s) == len(schema) for s in secrets])


def format_check(secrets: List[List[ShareValueType]],
                 schema: Sequence[Union[str, Schema]]) -> bool:
    # 存在チェック
    if not (schema and secrets):
        logger.error("Schema or secrets table are not exists.")
        return False
    # 重複チェック
    if not FormatChecker.check_duplicate(schema):
        logger.error("Duplicate schema name.")
        return False
    # サイズチェック
    if not FormatChecker.check_size(secrets, schema):
        logger.error("schema size and table colummn size are different.")
        return False
    return True


def to_float(val: str) -> float:
    """ If val is a float, convert as is; if it is a string, hash it. """
    try:
        return float(val)
    except ValueError:
        # k,m are constants used in the comparison operation
        # Due to the limitation of comparison operation,
        # k bits are taken out and divided by 2^m.
        k: int = 48
        m: int = 20
        hs: str = sha512(val.encode()).hexdigest()
        val_int: int = int(hs[:(k >> 2)], 16)
        val_float: float = val_int / pow(2, m)
        return val_float


def to_int(val: str, encoding='utf-8') -> int:
    encoded = val.encode(encoding)
    return int.from_bytes(encoded, byteorder='big')


def check_float_data(val: str) -> bool:
    try:
        _ = float(val)
        return True
    except ValueError:
        return False


def find_type(col_schema: str,
              col_data: List[str], is_matching_column: bool) \
        -> ShareValueTypeEnum.ValueType:
    # check tag
    type_str, *remains = col_schema.split(':')

    if len(remains) > 0:
        if type_str in SUPPORT_TAGS:
            return SUPPORT_TAGS[type_str]

    # check if the column of interest is the one used for matching
    if is_matching_column:
        return ShareValueTypeEnum.SHARE_VALUE_TYPE_FIXED_POINT

    # if even one string is not considered a floating point value,
    # treat the column as string data
    res_type = ShareValueTypeEnum.SHARE_VALUE_TYPE_FIXED_POINT

    if not all([check_float_data(v) for v in col_data]):
        res_type = \
            ShareValueTypeEnum.SHARE_VALUE_TYPE_UTF_8_INTEGER_REPRESENTATION

    return res_type


def find_types(schema: List[str],
               data: List[List[str]],
               matching_column: Optional[int] = None
               ) -> List[ShareValueTypeEnum.ValueType]:
    # transpose to get column oriented list
    transposed: List[List[str]] = np.array(
        data, dtype=str).transpose().tolist()
    return [find_type(sch, col, idx == matching_column)
            for idx, (sch, col) in enumerate(zip(schema, transposed), start=1)]


def convert(element: str,
            type_info: ShareValueTypeEnum.ValueType) -> ShareValueType:
    if type_info == ShareValueTypeEnum.Value('SHARE_VALUE_TYPE_FIXED_POINT'):
        return to_float(element)
    if type_info == ShareValueTypeEnum.Value(
            'SHARE_VALUE_TYPE_UTF_8_INTEGER_REPRESENTATION'):
        return to_int(element)
    return to_float(element)


def parse(data: List[List[str]], matching_column: Optional[int] = None) \
        -> Tuple[List[List[ShareValueType]], List[Schema]]:
    schema_name: List[str] = data[0]
    types = find_types(schema_name, data[1:], matching_column)
    schema = [Schema(name=name, type=type)
              for name, type in zip(schema_name, types)]

    # check size first because an iterator which `zip` bultin function returns
    # stops when the shortest iterable is exhausted
    if not FormatChecker.check_size(data[1:], schema):
        logger.error("schema size and table colummn size are different.")
        raise RuntimeError("規定されたフォーマットでないデータです．")

    secrets: List[List[ShareValueType]] = [
        [convert(x, t) for x, t in zip(row, types)] for row in data[1:]]

    if not format_check(secrets, schema):
        raise RuntimeError("規定されたフォーマットでないデータです．")

    return secrets, schema


def parse_to_bitvector(data: List[List[str]],
                       exclude: List[int] = [],
                       matching_column: Optional[int] = None) \
        -> Tuple[List[List[ShareValueType]], List[Schema]]:
    secrets, schema = parse(data, matching_column)

    secrets_bitbevtor: List[List[float]] = []
    schema_bitvector: List[Schema] = []
    for col, (sec, sch) in enumerate(zip(np.transpose(secrets), schema)):
        # 列が除外リストに含まれていたらそのままappend
        if col in exclude:
            secrets_bitbevtor.append(sec)
            schema_bitvector.append(
                Schema(
                    name=sch.name + "#0",
                    type=sch.type))
            continue

        # 座標圧縮
        position: dict = {}
        it: int = 0
        for key in sec:
            if key not in position:
                position[key] = it
                it += 1

        # bitvector化
        for key, val in position.items():
            bitvector: list = [1 if (key == k) else 0 for k in sec]
            sch_val: str = sch.name + "#" + str(val)
            secrets_bitbevtor.append(bitvector)
            schema_bitvector.append(
                Schema(
                    name=sch_val,
                    type=ShareValueTypeEnum.SHARE_VALUE_TYPE_FIXED_POINT))

    return np.transpose(secrets_bitbevtor).tolist(), schema_bitvector


def parse_csv(
    filename: str, matching_column: Optional[int] = None) \
        -> Tuple[List[List[ShareValueType]], List[Schema]]:
    with open(filename) as f:
        reader = csv.reader(f)
        text: List[List[str]] = [row for row in reader]
        return parse(text, matching_column)


def parse_csv_to_bitvector(filename: str,
                           exclude: List[int] = [],
                           matching_column: Optional[int] = None) ->  \
        Tuple[List[List[ShareValueType]], List[Schema]]:
    with open(filename) as f:
        reader = csv.reader(f)
        text: List[List[str]] = [row for row in reader]
        return parse_to_bitvector(text, exclude, matching_column)



from typing import Any, Dict

import pytest

from quickmpc.exception import ArgumentError, QMPCJobError, QMPCServerError
from quickmpc.qmpc_server import QMPCServer


def send_share_param(secrets=[[1, 2, 3]],
                     schema=["attr1", "attr2", "attr3"],
                     matching_column=1,
                     piece_size=1_000_000):
    return (secrets, schema, matching_column, piece_size)


def execute_computation_param(method_id=1,
                              data_ids=["data_id1", "data_id2"],
                              join=[1],
                              index=[1, 1],
                              src=[0],
                              target=[1]):
    return (method_id, (data_ids, join, index), (src, target))


class TestQMPC:
    qmpc: QMPCServer = QMPCServer(
        ["http://localhost:50001",
         "http://localhost:50002",
         "http://localhost:50003"],
        "token_demo"
    )

    @pytest.mark.parametrize(
        ("params"), [
            (send_share_param()),
            (send_share_param(secrets=[[1], [2], [3]], schema=["a1"])),
            (send_share_param(secrets=[[0, 0, 0]])),
            (send_share_param(secrets=[[1e10, 1e10, 1e10]])),
            (send_share_param(secrets=[[-1e10, -1e10, -1e10]])),
            (send_share_param(secrets=[[1e-10, 1e-10, 1e-10]])),
            (send_share_param(secrets=[[-1e-10, -1e-10, -1e-10]])),
        ]
    )
    def test_send_shares(self, params,
                         run_server1, run_server2, run_server3):
        """ serverにシェアを送れるかのTest"""
        response: Dict[str, Any] = self.qmpc.send_share(*params)
        assert (response["is_ok"])

    @pytest.mark.parametrize(
        ("params", "expected_exception"), [
            # piece_sizeが範囲外
            (send_share_param(piece_size=500), RuntimeError),
            (send_share_param(piece_size=10000000), RuntimeError),
            # matching_columnが範囲外
            (send_share_param(matching_column=-1), RuntimeError),
            (send_share_param(matching_column=4), RuntimeError),
            # empty
            (send_share_param(secrets=[]), TypeError),
            # schemaに同じものが含まれる
            (send_share_param(schema=["a1", "a1", "a2"]), RuntimeError),
            # 列数が異なる
            (send_share_param(secrets=[[1, 1, 2], [2, 3]]), RuntimeError),
            # 文字列が含まれる
            (send_share_param(secrets=[["a", 2, 3]]), ArgumentError),
            # Noneが含まれる
            (send_share_param(secrets=[[None, 2, 3]]), TypeError),
        ]
    )
    def test_send_shares_errorhandring(self, params, expected_exception,
                                       run_server1, run_server2, run_server3):
        with pytest.raises(expected_exception):
            self.qmpc.send_share(*params)

    def test_delete_shares(self, run_server1, run_server2, run_server3):
        """ serverにシェア削除要求を送れるかのTest"""
        response: Dict[str, Any] = self.qmpc.delete_share([])
        assert (response["is_ok"])

    @pytest.mark.parametrize(
        ("params"), [
            execute_computation_param(method_id=1, target=[]),
            execute_computation_param(method_id=2, target=[]),
            execute_computation_param(method_id=3, target=[]),
            execute_computation_param(method_id=4),
            execute_computation_param(method_id=5),
            execute_computation_param(method_id=6),
        ]
    )
    def test_execute_computation(self, params,
                                 run_server1, run_server2, run_server3):
        """ serverに計算リクエストを送れるかのTest"""
        response: Dict[str, Any] = self.qmpc.execute_computation(*params)
        assert (response["is_ok"])

    @pytest.mark.parametrize(
        ("params", "expected_exception"), [
            # data_idsの要素数-1とjoinの要素数が一致していない
            (execute_computation_param(data_ids=["id1", "id2"], join=[1, 1]),
             ArgumentError),
            # data_idsの要素数とindexの要素数が一致していない
            (execute_computation_param(data_ids=["id1", "id2"], index=[1]),
             ArgumentError),
            # joinの値が範囲外
            (execute_computation_param(join=[-1]),
             ArgumentError),
            (execute_computation_param(join=[3]),
             ArgumentError),
            # QMPCJobErrorとして例外がthrowされるか
            (execute_computation_param(src=[1000000000]),
             QMPCJobError),
            # QMPCServerErrorとして例外がthrowされるか
            (execute_computation_param(data_ids=["UnregisteredDataId", "id2"]),
             QMPCServerError),
        ]
    )
    def test_execute_computation_errorhandring(self, params,
                                               expected_exception,
                                               run_server1,
                                               run_server2,
                                               run_server3):
        with pytest.raises(expected_exception):
            self.qmpc.execute_computation(*params)

    def test_get_computation_resultRequest(self, run_server1,
                                           run_server2, run_server3):
        """ serverから結果を得られるかのTest """
        job_uuid: str = "uuid"
        response: Dict[str, Any] = self.qmpc.get_computation_result(
            job_uuid, None)
        assert (response["is_ok"])

    def test_get_data_list(self, run_server1, run_server2, run_server3):
        """ serverにシェアを送れるかのTest"""
        response: Dict[str, Any] = self.qmpc.get_data_list()
        assert (response["is_ok"])

    def test_get_job_error_info(self, run_server1, run_server2, run_server3):
        # 例外がthrowされないことをテスト
        response: Dict[str, Any] = self.qmpc.get_job_error_info("test")

        assert (response["is_ok"])
        for res in response["job_error_info"]:
            assert (res.what == "QMPCJobError")

    qmpc_failed: QMPCServer = QMPCServer(
        # 通信失敗する場合をテストする用のサーバー
        ["http://localhost:50011",
         "http://localhost:50012",
         "http://localhost:50013"],
        "token_demo",
        10,
        0
    )

    @pytest.mark.parametrize(
        ("function", "argument"), [
            (qmpc_failed.send_share, send_share_param()),
            (qmpc_failed.delete_share, [[]]),
            (qmpc_failed.execute_computation, execute_computation_param()),
            (qmpc_failed.get_data_list, []),
            (qmpc_failed.get_elapsed_time, ["uuid"]),
            (qmpc_failed.get_computation_result, ["uuid", None]),
            (qmpc_failed.get_job_error_info, ["uuid"]),
        ]
    )
    def test_retry(self, function, argument, caplog,
                   run_server1, run_server2, run_server3):
        # 10回の retry に失敗したら "All 10 times it was an error" が log に出るかをテスト
        _ = function(*argument)
        assert "All 10 times it was an error" in caplog.text
